import itertools
import json
import asyncio
import websockets

from pseuco_parser import createParser, UnsupportedPseuCoException
from semantic import Z3Compiler, seqconsist_axioms, weak_order_axioms


def parseProgram(program):
    ret = {}
    counter = 0
    for agent_name in itertools.chain.from_iterable(
            program.startedAgentNames.asList()):
        agent = program.initialization.agent_declarations[agent_name]
        body = agent.functionBody
        cfg_predecessors = []
        cfg_graph_represantation = []
        for statement in body.statements:
            cfg_graph_represantation += statement.create_node_data(
                cfg_predecessors)
            cfg_predecessors = statement.get_cfg_exit_nodes()
        ret[agent_name + str(counter)] = cfg_graph_represantation
        counter += 1
    return ret


async def parse(websocket, path):
    program = None
    reply = None
    compiler = None
    # TODO: use websocket.remote_address for bookkeeping
    async for message in websocket:
        try:
            decoded_message = json.loads(message)
            if decoded_message["type"] == "code":
                tinyPseucoParser = createParser()
                program = tinyPseucoParser.parseString(decoded_message["code"])
                reply = {}
                reply["cfg"] = parseProgram(program)
                reply["type"] = "parseResult"
                compiler = Z3Compiler()
                compiler.compile(program)
            elif decoded_message["type"] == "listAllReads":
                reply = {}
                reply["type"] = "allReads"
                if decoded_message["memoryModel"] == "seqConsist":
                    reply["allReads"] = compiler.check_sat(seqconsist_axioms)
                elif decoded_message["memoryModel"] == "weak":
                    reply["allReads"] = compiler.check_sat(weak_order_axioms)
        except UnsupportedPseuCoException as e:
            reply = {
                "type": "error",
                "errorMessage": str(e),
            }
        await websocket.send(json.dumps(reply))


if __name__ == "__main__":

    start_server = websockets.server.serve(parse, '127.0.0.1', 5678)

    loop = asyncio.get_event_loop()

    server = loop.run_until_complete(start_server)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.run_until_complete(server.wait_closed())
        # loop.close() TODO: shouldn't we need to call close ?!?
